//============================================================================
//  aPCI-8102 for PCI Linux Device Driver
//  apci8102.c
//  Author   K. Fujimaki(ADTEK SYSTEM SCIENCE Co.,Ltd.)
//  Date     2001/07/16
//  Rev.     1.0.0
//============================================================================

#ifndef MODULE
#define MODULE
#endif // MODULE

#ifndef __KERNEL__
#define __KERNEL__
#endif // __KERNEL__

#include <linux/autoconf.h>
#if defined(CONFIG_MODVERSIONS) && !defined(MODVERSIONS)
#	define MODVERSIONS
#endif

#ifdef MODVERSIONS
#	include <linux/modversions.h>
#endif

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/fs.h>		// used file oprations
#include <linux/pci.h>		// used pci functions
#include <asm/io.h>		// used inb, outb
#include <asm/uaccess.h>	// used verify_area
#include <linux/poll.h>		// used poll_wait
#include <asm/delay.h>		// used udelay

#include "apci8102.h"

/*
 * Split minors in two parts
 */
#define TYPE(dev)		(MINOR(dev) >> 4)	// high nibble
#define NUM(dev)		(MINOR(dev) & 0xf)	// low	nibble


#define APCI8102_NAME		"apci8102"

#define APCI8102_VENDOR_ID	0x136C		// ǥХID,٥ID פʤ
#define APCI8102_DEVICE_ID	0x8102		// ɥ饤ФϿǤʤΤ

#define APCI8102_CTL_PORT	0x04		// ȥݡȤΥɥ쥹

#ifdef APCI8102_DEBUG
#	define APCI8102_DELAY_CLOSE	1000 	// Ԥ
#else
#	define APCI8102_DELAY_CLOSE	200 	// Ԥ
#endif

#define PWARNING(fmt, args...)	printk(KERN_WARNING "[APCI8102]: " fmt, ## args)

/*----------------- Define Device Information ------------------------------*/

typedef struct _devinf
{
	unsigned int		io_addr;	// I/O	  Base Address
	unsigned long		f_version;	// ե뼱̻
} devinf, *pdevinf;

static devinf		devctl[APCI8102_MAX_DEVICES];	// ǥХ
static unsigned int	uapci8102major;			// Major Device Number

static int devmajor = 0;


/*----------------- File Operations ----------------------------------------*/

static		int apci8102_ioctl(struct inode*, struct file*, unsigned int, unsigned long);
static		int apci8102_open(struct inode* inode, struct file* file);
static		int apci8102_close(struct inode* inode, struct file* file);

static struct file_operations apci8102_fops =
{
	NULL,		// loff_t	llseek(struct file*, loff_t, int)
	NULL,		// ssize_t	read(struct file*, char*, size_t, loff_t*)
	NULL,		// ssize_t	write(struct file*, const char*, size_t, loff_t*)
	NULL,		// int		readdir(struct file*, void*, filldir_t)
	NULL,		// unsigned int poll(struct file*, struct poll_table_struct*)
	apci8102_ioctl,	// int		ioctl(struct inode*, struct file*, unsigned int, unsigned long)
	NULL,		// int		mmap(struct file*, struct vm_area_struct*)
	apci8102_open,	// int		open(struct inode*, struct file*)
	NULL,		// int		flush(struct file*)
	apci8102_close,	// int		release(struct inode*, struct file*)
	NULL,		// int		fsync(struct file*, struct dentry*)
	NULL,		// int		fasync(int, struct file*, int)
	NULL,		// int		check_media_change(kdev_t dev)
	NULL,		// int		revalidate(kdev_t dev)
	NULL,		// int		lock(struct file*, int, struct file_lock*)
};


/*----------------- Define Static Functions --------------------------------*/

static int		init_driver_info(void);
static int		setup_driver(void);
static void		release_driver(void);
static int		get_device(void);
static unsigned int	get_device_base_addr(unsigned int base_addr);
static int		regist_char_driver(void);
static int		unregist_char_driver(void);
static int		get_device_state(int ref);

static int	verify_arg(unsigned int cmd, unsigned long arg);
static int	apci8102_ioc_get_dirs(struct file* file, unsigned long arg);
static int	apci8102_ioc_set_dirs(struct file* file, unsigned long arg);
static int	apci8102_ioc_inp(struct file* file, unsigned long arg);
static int	apci8102_ioc_outp(struct file* file, unsigned long arg);

static int	is_port(unsigned port);


MODULE_PARM(devmajor, "i");	// ᥸㡼ֹѥ᡼


/*==========================================================================*/
/*  init_module - ɥ饤ФϿ					    */
/*==========================================================================*/
int init_module(void)
{
	int result;

	// PCI ǽ̵ͭȽ
	if (!pci_present())
		return -ENODEV;

	// ǥХɥ饤ѿν
	result = init_driver_info();
	if (result < 0)
		return result;

	// ǥХɥ饤ФΥåȥå
	result = setup_driver();
	if (result < 0)
		return result;

	return 0;
}

/*==========================================================================*/
/*  cleanup_module - ɥ饤ФϿä				    */
/*==========================================================================*/
void cleanup_module(void)
{
	// ǥХɥ饤Фβ
	release_driver();
}

/*--------------------------------------------------------------------------*/
/*  apci8102_ioctl - ǥХͭ					    */
/*--------------------------------------------------------------------------*/
static int apci8102_ioctl(struct inode* inode, struct file* file, unsigned int cmd, unsigned long arg)
{
	int err = verify_arg(cmd, arg);
	if (err)
		return err;

	switch (cmd)
	{
		case APCI8102_GET_DIRS:
			return apci8102_ioc_get_dirs(file, arg);

		case APCI8102_SET_DIRS:
			return apci8102_ioc_set_dirs(file, arg);

		case APCI8102_IN_PORT:
			return apci8102_ioc_inp(file, arg);

		case APCI8102_OUT_PORT:
			return apci8102_ioc_outp(file, arg);

	}

	return 0;
}

/*--------------------------------------------------------------------------*/
/*  apci8102_open - ǥХΥץ					    */
/*--------------------------------------------------------------------------*/
static int apci8102_open(struct inode* inode, struct file* file)
{
	int ref;
	int result;

	PDEBUG("Open !! %ld\n", file->f_version);

	// Ȥ
	ref = NUM(inode->i_rdev);

	PDEBUG("ref index=%d\n", ref);

	// ǥХξ
	result = get_device_state(ref);
	if (result < 0)
		return result;

	// Ѿ֤ˤ
	devctl[ref].f_version = file->f_version;

	// ǡȰ¸
	file->private_data = (void*)ref;

	MOD_INC_USE_COUNT;		// ȥ+

	return 0;
}

/*--------------------------------------------------------------------------*/
/*  apci8102_close - ǥХΥ					    */
/*--------------------------------------------------------------------------*/
static int apci8102_close(struct inode* inode, struct file* file)
{
	int ref;

	PDEBUG("Close !!\n");

	// ǡλȰ
	ref = (int)file->private_data;

	// ̤Ѿ֤ˤ
	devctl[ref].f_version = 0;

	MOD_DEC_USE_COUNT;		// ȥ-

	return 0;
}

/*--------------------------------------------------------------------------*/
/*  init_driver_info - ǥХɥ饤дѿν			    */
/*  Input : none							    */
/*  Output: return - 0 					    */
/*--------------------------------------------------------------------------*/
static int init_driver_info(void)
{
	int i;

	for (i = 0; i < APCI8102_MAX_DEVICES; i++)
	{
		devctl[i].io_addr	= 0x0000;
		devctl[i].f_version = 0;
	}

	return 0;
}

/*--------------------------------------------------------------------------*/
/*  setup_driver - ǥХɥ饤ФΥåȥå			    */
/*  Input : none							    */
/*  Output: return - 0 ʳ åȥå׼				    */
/*  		     0 åȥå					    */
/*--------------------------------------------------------------------------*/
static int setup_driver(void)
{
	int result;

	result = get_device();
	if (result < 0)
		return result;

	// character device driver registration
	result = regist_char_driver();
	if (result < 0)
		return result;

	return 0;
}

/*--------------------------------------------------------------------------*/
/*  release_driver -  ǥХɥ饤Фβ				    */
/*  Input : none							    */
/*  Output: none							    */
/*--------------------------------------------------------------------------*/
static void release_driver(void)
{
	// character device driver unregistration
	unregist_char_driver();
}

/*--------------------------------------------------------------------------*/
/*  get_device - 󥹥ȡ뤵ƤǥХ		    */
/*  Input : none							    */
/*  Output: return - -ENODEV ǥХϸդʤ			    */
/*  		     0	     ǥХϸդä			    */
/*--------------------------------------------------------------------------*/
static int get_device(void)
{
	struct pci_dev* pdev = NULL;

	unsigned int  io_addr;		// I/O Base Address

	int i;

	for (i = 0; i < APCI8102_MAX_DEVICES; i++)
	{
		// PCI ǥХ
		pdev = pci_find_device(APCI8102_VENDOR_ID,
					   APCI8102_DEVICE_ID, pdev);
		if (pdev == NULL)
			break;

		//  I/O ١ɥ쥹
		io_addr	  = get_device_base_addr(pdev->base_address[2]);

		PDEBUG("io_addr:   %X\n", io_addr);

		// ǡ
		devctl[i].io_addr   = io_addr;	// I/O	  Base Address save
	}

	if (i == 0)
	{
		PWARNING("Not found Device(s)\n");
		return -ENODEV;
	}

	return 0;
}

/*--------------------------------------------------------------------------*/
/*  get_device_base_addr - ѲǽʥǥХΥ١ɥ쥹	    */
/*  Input : base_addr - ١ɥ쥹					    */
/*  Output: return - Խ١ɥ쥹				    */
/*--------------------------------------------------------------------------*/
static unsigned int get_device_base_addr(unsigned int base_addr)
{
	// I/O ١ɥ쥹쥸
	if ((base_addr & PCI_BASE_ADDRESS_SPACE) == PCI_BASE_ADDRESS_SPACE_IO)
		return base_addr & PCI_BASE_ADDRESS_IO_MASK;

	// Memory ١ɥ쥹쥸
	return base_addr & PCI_BASE_ADDRESS_MEM_MASK;
}

/*--------------------------------------------------------------------------*/
/*  regist_char_driver - ᥸㡼ֹưŪƤȥɥ饤ФϿ	    */
/*  Input : none							    */
/*  Output: return - register_chrdev 				    */
/*  		     0	Ͽ					    */
/*--------------------------------------------------------------------------*/
static int regist_char_driver(void)
{
	int result;

	PDEBUG("devmajor=%d\n", devmajor);

	if (devmajor == 0)
	{
		// ᥸㡼ֹ椬ꤵƤʤ
		// ưŪ᥸㡼ֹμ
		result = register_chrdev(0, APCI8102_NAME, NULL);
		if (result < 0)
		{
			PDEBUG("get major device number faild.\n");
			return result;
		}

		uapci8102major = result;
	}
	else
	{
		// ᥸㡼ֹ椬ꤵƤ
		uapci8102major = devmajor;
	}

	PDEBUG("major number=%u\n", uapci8102major);

	// 饯ǥХɥ饤ФϿ
	result = register_chrdev(uapci8102major, APCI8102_NAME, &apci8102_fops);
	if (result < 0)
	{
		PDEBUG("register_chrdev faild.\n");
		return result;
	}

	return 0;
}

/*--------------------------------------------------------------------------*/
/*  unregist_char_driver - ɥ饤ФϿä			    */
/*  Input : none							    */
/*  Output: return - unregister_chrdev 				    */
/*  		     0	Ͽ					    */
/*--------------------------------------------------------------------------*/
static int unregist_char_driver(void)
{
	int result;

	// 饯ǥХɥ饤ФϿ
	result = unregister_chrdev(uapci8102major, APCI8102_NAME);
	if (result < 0)
	{
		PDEBUG("unregister_chrdev faild.\n");
		return result;
	}

	return 0;
}

/*--------------------------------------------------------------------------*/
/*  get_device_state - ǥХξ֤				    */
/*  Input : ref	  - Ȱ						    */
/*  Output: return - -ENODEV ΰ֤˥ǥХϤʤ			    */
/*  		     -EBUSY  					    */
/*  		      0      Ѳǽ 					    */
/*  Notes : ref >= 0 - ref <= 15 ޤǤʤ				    */
/*--------------------------------------------------------------------------*/
static int get_device_state(int ref)
{
	if (devctl[ref].io_addr == 0x0000)
		return -ENODEV;

	if (devctl[ref].f_version != 0)
		return -EBUSY;

	return 0;
}

/*--------------------------------------------------------------------------*/
/*  verify_arg - ǥХͭΰ				    */
/*  Input : cmd - ioctl 楳ޥ					    */
/*  	    arg - 桼֤ؤΥɥ쥹				    */
/*  Output: return - -EINVAL ̵ʰ					    */
/*  		     verify_area 				    */
/*  		     0		 Ѳǽ 				    */
/*--------------------------------------------------------------------------*/
static int verify_arg(unsigned int cmd, unsigned long arg)
{
	int size = _IOC_SIZE(cmd);

	/* don't even decode wrong cmds: better returning  EINVAL than EFAULT */
	if (_IOC_TYPE(cmd) != APCI8102_IOC_MAGIC)
		return -EINVAL;

	if (_IOC_NR(cmd) > APCI8102_IOC_MAXNR)
		return -EINVAL;

	/*
	 * the type is a bitmask, and VERIFY_WRITE catches R/W
	 * transfers. Note that the type is user-oriented, while
	 * verify_area is kernel-oriented, so the concept of "read" and
	 * "write" is reversed
	 */
	if (_IOC_TYPE(cmd) & _IOC_READ)
		return verify_area(VERIFY_WRITE, (void*)arg, size);

	if (_IOC_TYPE(cmd) & _IOC_WRITE)
		return verify_area(VERIFY_READ, (void*)arg, size);

	return 0;
}

/*--------------------------------------------------------------------------*/
/*  apci8102_ioc_get_dirs - ǥ쥯			    */
/*  Input : file - ͥ file ¤ΤΥݥ			    */
/*  	    arg	 - 桼֤ؤΥɥ쥹				    */
/*  Output: return - -EINVAL ̵ʰ					    */
/*  		     0       ｪλ					    */
/*--------------------------------------------------------------------------*/
static int apci8102_ioc_get_dirs(struct file* file, unsigned long arg)
{
	int ref;
	unsigned char dirs;

	PDEBUG("Get Directions\n");

	// ǡλȰ
	ref = (int)file->private_data;

	// ǥ쥯
	dirs = ~(inb(devctl[ref].io_addr + APCI8102_CTL_PORT)) & 0x0f;

	// 桼֤إǡ
	copy_to_user((void*)arg, &dirs, sizeof(unsigned char));

	return 0;
}

/*--------------------------------------------------------------------------*/
/*  apci8102_ioc_set_dirs - ǥ쥯				    */
/*  Input : file - ͥ file ¤ΤΥݥ			    */
/*  	    arg	 - 桼֤ؤΥɥ쥹				    */
/*  Output: return - -EINVAL ̵ʰ					    */
/*  		     0	     ｪλ					    */
/*--------------------------------------------------------------------------*/
static int apci8102_ioc_set_dirs(struct file* file, unsigned long arg)
{
	int ref;
	unsigned char dirs;

	PDEBUG("Get Directions\n");

	// ǡλȰ
	ref = (int)file->private_data;

	// 桼֤ǡ
	copy_from_user(&dirs, (void*)arg, sizeof(unsigned char));

	// ǥ쥯
	outb(~dirs | 0xf0, devctl[ref].io_addr + APCI8102_CTL_PORT);

	return 0;
}

/*--------------------------------------------------------------------------*/
/*  apci8102_ioc_inp - ϥݡȤǡ			    */
/*  Input : file - ͥ file ¤ΤΥݥ			    */
/*  	    arg	 - 桼֤ؤΥɥ쥹				    */
/*  Output: return - -EINVAL ̵ʰ					    */
/*  		      0      ｪλ					    */
/*--------------------------------------------------------------------------*/
static int apci8102_ioc_inp(struct file* file, unsigned long arg)
{
	int ref;
	apci8102_io data;

	PDEBUG("In Port\n");

	// ǡλȰ
	ref = (int)file->private_data;

	// 桼֤ǡ
	copy_from_user(&data, (void*)arg, sizeof(data));

	// ݡȳǧ
	if (!is_port(data.port))
		return -EINVAL;

	data.value = inb(devctl[ref].io_addr + data.port);

	// 桼֤إǡ
	copy_to_user((void*)arg, &data, sizeof(data));

	return 0;
}

/*--------------------------------------------------------------------------*/
/*  apci8102_ioc_outp - ϥݡȤإǡ			    */
/*  Input : file - ͥ file ¤ΤΥݥ			    */
/*  	    arg	 - 桼֤ؤΥɥ쥹				    */
/*  Output: return - -EINVAL ̵ʰ					    */
/*  		     0	     ｪλ					    */
/*--------------------------------------------------------------------------*/
static int apci8102_ioc_outp(struct file* file, unsigned long arg)
{
	int ref;
	apci8102_io data;

	PDEBUG("Out Port\n");

	// ǡλȰ
	ref = (int)file->private_data;

	// 桼֤ǡ
	copy_from_user(&data, (void*)arg, sizeof(data));

	// ݡȳǧ
	if (!is_port(data.port))
		return -EINVAL;

	outb(data.value, devctl[ref].io_addr + data.port);

	return 0;
}

/*--------------------------------------------------------------------------*/
/*  is_port - ݡȤ					    */
/*  Input : port - Ĵ٤ݡ					    */
/*  Output: return - 0 ̵						    */
/*  		     1 						    */
/*--------------------------------------------------------------------------*/
static int is_port(unsigned int port)
{
	if (port >= APCI8102_MAX_PORTS)
		return 0;

	return 1;
}

